/**
 * Copyright (C) 2007-2014 Lawrence Murray
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * @author Lawrence Murray <lawrence@indii.org>
 * $Rev$
 * $Date$
 */
#include "ClusterModel.hpp"

#include "ClusterModelObserver.hpp"

#include <cassert>
#include <ctime>
#include <limits>

namespace ublas = boost::numeric::ublas;

using namespace indii;
using namespace std;

ClusterModel::ClusterModel(ImageResource* res) :
    Model(res), clusterer(time(NULL)), locked(false) {
  this->saturationThreshold = 0.0625f;
  this->maxPixels = 50000;
  this->reps = 12;
  this->k = 4;
  this->hard = 0.9f;
  this->saturationDecay = 0.0625f;
  this->centroidDecay = 0.0f;
  this->saturationSoftness = 0.125f;
  this->centroidSoftness = 0.0f;
  this->colours.resize(k);
  this->mus.resize(k);
  this->hues.resize(k, 0.0f);
  this->sats.resize(k, -1.0f);
  this->lights.resize(k, 0.0f);
  this->alphas.resize(k, 1.0f);
}

ClusterModel::~ClusterModel() {
  //
}

void ClusterModel::setDefaults() {
  lock();
  setHard(0.9f);
  setSaturationThreshold(0.0625f);
  setMaxPixels(50000);
  setNumRepetitions(12);
  setNumClusters(4);
  setSaturationDecay(0.0625f);
  setCentroidDecay(0.0f);
  setSaturationSoftness(0.125f);
  setCentroidSoftness(0.0f);
  setChannelMix(0.299f, 0.587f, 0.114f);
  setForDialog();
  unlock();
  prepare();
  cluster();

  notifyAll();
  notifyNumClustersChange();
}

void ClusterModel::setForDialog() {
  fill(hues.begin(), hues.end(), 0.0f);
  fill(sats.begin(), sats.end(), -1.0f);
  fill(lights.begin(), lights.end(), 0.0f);
  fill(alphas.begin(), alphas.end(), 1.0f);
}

void ClusterModel::setNumClusters(const int k) {
  if (this->k != k) {
    this->k = k;
    colours.resize(k, *wxWHITE);
    mus.resize(k);
    hues.resize(k);
    sats.resize(k);
    lights.resize(k);

    fill(sats.begin(), sats.end(), -1.0f);
    fill(hues.begin(), hues.end(), 0.0f);
    fill(lights.begin(), lights.end(), 0.0f);
    fill(alphas.begin(), alphas.end(), 1.0f);

    cluster();
    notifyAll();
    notifyNumClustersChange();
  }
}

void ClusterModel::setHard(const float hard) {
  float hard1 = bound(0.0f, hard, 1.0f);
  if (this->hard != hard1) {
    this->hard = hard1;
    notifyAll();
    notifyIsHardChange();
  }
}

void ClusterModel::setNumRepetitions(const int reps) {
  int reps1 = bound(1u, reps, 20u);
  if (this->reps != reps1) {
    this->reps = reps1;
    notifyAll();
    notifyNumRepetitionsChange();
  }
}

void ClusterModel::setSaturationThreshold(const float x) {
  float x1 = bound(0.0f, x, 1.0f);
  if (this->saturationThreshold != x1) {
    this->saturationThreshold = x1;
    prepare();
    notifyAll();
    notifySaturationThresholdChange();
  }
}

void ClusterModel::setMaxPixels(const int n) {
  int n1 = bound(1000, n, 1e6);
  if (this->maxPixels != n1) {
    this->maxPixels = n1;
    prepare();
    notifyAll();
    notifyMaxPixelsChange();
  }
}

void ClusterModel::setSaturationDecay(const float x) {
  float x1 = bound(0.0f, x, 1.0f);
  if (this->saturationDecay != x1) {
    this->saturationDecay = x1;
    notifyAll();
    notifySaturationDecayChange();
  }
}

void ClusterModel::setCentroidDecay(const float x) {
  float x1 = bound(0.0f, x, 1.0f);
  if (this->centroidDecay != x1) {
    this->centroidDecay = x1;
    notifyAll();
    notifyCentroidDecayChange();
  }
}

void ClusterModel::setSaturationSoftness(const float x) {
  float x1 = bound(0.0f, x, 1.0f);
  if (this->saturationSoftness != x1) {
    this->saturationSoftness = x1;
    notifyAll();
    notifySaturationSoftnessChange();
  }
}

void ClusterModel::setCentroidSoftness(const float x) {
  float x1 = bound(0.0f, x, 1.0f);
  if (this->centroidSoftness != x1) {
    this->centroidSoftness = x1;
    notifyAll();
    notifyCentroidSoftnessChange();
  }
}

void ClusterModel::setColour(const int i, const wxColour& col) {
  if (i < k) {
    value_type x(3);
    x(0) = col.Red() / 255.0f;
    x(1) = col.Green() / 255.0f;
    x(2) = col.Blue() / 255.0f;
    distance_type::prepare(x);
    colours[i] = col;
    mus[i] = x;
  }
}

void ClusterModel::setColour(const int i, const float r, const float g, const float b) {
  if (i < k) {
    //float r1 = bound(0.0f, r, 1.0f);
    //float g1 = bound(0.0f, g, 1.0f);
    //float b1 = bound(0.0f, b, 1.0f);

    value_type x(3);
    x(0) = r;
    x(1) = g;
    x(2) = b;
    //distance_type::prepare(x);

    wxColour col;
    col.Set(255.0f*r, 255.0f*g, 255.0f*b);
  
    colours[i] = col;
    mus[i] = x;
  }
}

void ClusterModel::setHue(const int i, const float x) {
  if (i < k) {
    float x1 = bound(-0.5f, x, 0.5f);
    if (hues[i] != x1) {
      hues[i] = x1;
      notifyAll();
      notifyHueChange(i);
    }
  }
}

void ClusterModel::setSat(const int i, const float x) {
  if (i < k) {
    float x1 = bound(-1.0f, x, 1.0f);
    if (sats[i] != x1) {
      sats[i] = x1;
      notifyAll();
      notifySatChange(i);
    }
  }
}

void ClusterModel::setLight(const int i, const float x) {
  if (i < k) {
    float x1 = bound(-1.0f, x, 1.0f);
    if (lights[i] != x1) {
      lights[i] = x1;
      notifyAll();
      notifyLightChange(i);
    }
  }
}

void ClusterModel::setAlpha(const int i, const float x) {
  if (i < k) {
    float x1 = bound(0.0f, x, 1.0f);
    if (alphas[i] != x1) {
      alphas[i] = x1;
      notifyAll();
      notifyAlphaChange(i);
    }
  }
}

void ClusterModel::show(const int i, const bool on) {
  if (i < k && isShown(i) != on) {
    if (on) {
      sats[i] = 0.0f;
    } else {
      sats[i] = -1.0f;
    }
    notifyAll();
    notifySatChange(i);
  }
}

void ClusterModel::showAll(const bool on) {
  int i;
  for (i = 0; i < k; i++) {
    show(i, on);
  }
}

void ClusterModel::calcFg(const wxRect& rect, const float scale, Image& o) {
  /* pre-conditions */
  assert(o.getWidth() == rect.width);
  assert(o.getHeight() == rect.height);

  Image& img = *res->get(scale);

#pragma omp parallel
  {
    ColourSpace::rgb_t rgb(3);
    ColourSpace::hsl_t hsl(3);
    std::vector<float> ds(k);

    int x, y, cluster;
    float decay, light, sat;

#pragma omp for
    for (y = 0; y < rect.height; ++y) {
      for (x = 0; x < rect.width; ++x) {
        rgb(0) = img.r(rect.y + y, rect.x + x);
        rgb(1) = img.g(rect.y + y, rect.x + x);
        rgb(2) = img.b(rect.y + y, rect.x + x);
        cs.rgb2hsl(rgb, hsl);
        decay = threshold(hsl(1), saturationDecay, saturationSoftness);
        
        /* cluster adjustments */
        distance_type::prepare(rgb);
        cluster = clusterer.assign(rgb, ds, mus);
        if (isHard()) {
          /* hue */
          hsl(0) = fmod(1.0f + hsl(0) + hues[cluster], 1.0f);

          /* saturation */
          sat = sats[cluster];
          if (sat >= 0.0f) {
            hsl(1) /= 1.0f - sat;
          } else {
            hsl(1) *= 1.0f + sat;
          }
          hsl(1) = bound(0.0f, decay*hsl(1), 1.0f);

          /* lightness */
          light = decay*lights[cluster];
          if (light >= 0.0f) {
            hsl(2) /= 1.0f - light;
          } else {
            hsl(2) *= 1.0f + light;
          }
          hsl(2) = bound(0.0f, hsl(2), 1.0f);

          /* alpha */
          //a = bound(0.0f, decay*alphas[cluster], 255.0f);
        } else {
          int i;
          float h = 0.0f, s = 0.0f, l = 0.0f/*, a = 0.0f*/, D = 0.0f;
          for (i = 0; i < k; ++i) {
            ds[i] = std::pow(1.0f - ds[i], 256.0f * hard);
            D += ds[i];
          }
          if (D >= 0.0f) {
            for (i = 0; i < k; ++i) {
              ds[i] /= D;
              if (ds[i] > 0.0f && ds[i] <= 1.0f) { // D >= 0.0 check may use extended precision, so ds[i] may not be finite
                h += ds[i] * hues[i];
                s += ds[i] * sats[i];
                l += ds[i] * lights[i];
              }
            }
          }
          //a += ds[i]*alphas[i];

          /* hue */
          hsl(0) = fmod(1.0f + hsl(0) + h, 1.0f);

          /* saturation */
          sat = s;
          if (sat >= 0.0f) {
            hsl(1) /= 1.0f - sat;
          } else {
            hsl(1) *= 1.0f + sat;
          }
          hsl(1) = bound(0.0f, decay*hsl(1), 1.0f);

          /* lightness */
          light = decay*l;
          if (light >= 0.0f) {
            hsl(2) /= 1.0f - light;
          } else {
            hsl(2) *= 1.0f + light;
          }
          hsl(2) = bound(0.0f, hsl(2), 1.0f);

          /* alpha */
          // already set
        }

        /* convert back */
        cs.hsl2rgb(hsl, rgb);
        o.r(y, x) = rgb(0);
        o.g(y, x) = rgb(1);
        o.b(y, x) = rgb(2);
        //o.a(y, x) = a;
      }
    }
  }
}

void ClusterModel::calcAlpha(const int i, const wxRect& rect,
    const float scale, channel& c) {
  /* pre-conditions */
  assert(i < k);
  assert((int)c.size1() == rect.height && (int)c.size2() == rect.width);

  Image* img = res->get(scale);

#pragma omp parallel default(shared)
  {
    ColourSpace::rgb_t rgb(3);
    ColourSpace::hsl_t hsl(3);
    value_type pixel(3);
    std::vector<float> ds(k);
    int x, y, cluster;

#pragma omp for
    for (y = 0; y < rect.height; ++y) {
      for (x = 0; x < rect.width; ++x) {
        rgb(0) = img->r(rect.y + y, rect.x + x);
        rgb(1) = img->g(rect.y + y, rect.x + x);
        rgb(2) = img->b(rect.y + y, rect.x + x);
        cs.rgb2hsl(rgb, hsl);
        pixel(0) = rgb(0);
        pixel(1) = rgb(1);
        pixel(2) = rgb(2);

        distance_type::prepare(pixel);
        cluster = clusterer.assign(pixel, ds, mus);
        if (cluster == i) {
          c(y, x) = threshold(hsl(1), saturationDecay, saturationSoftness);
        } else {
          c(y, x) = 0;
        }
      }
    }
  }
}

sparse_mask ClusterModel::calcMask(const int i, const wxRect& rect,
    const float scale) {
  /* pre-condition */
  assert(i < k);

  sparse_mask m(rect.height, rect.width);
  Image* img = res->get(scale);

#pragma omp parallel
  {
    ColourSpace::rgb_t rgb(3);
    ColourSpace::hsl_t hsl(3);
    value_type pixel(3);
    std::vector<float> ds(k);
    int x, y, cluster;

#pragma omp for
    for (y = 0; y < rect.height; ++y) {
      for (x = 0; x < rect.width; ++x) {
        rgb(0) = img->r(rect.y + y, rect.x + x);
        rgb(1) = img->g(rect.y + y, rect.x + x);
        rgb(2) = img->b(rect.y + y, rect.x + x);
        cs.rgb2hsl(rgb, hsl);
        pixel = rgb;

        distance_type::prepare(pixel);
        cluster = clusterer.assign(pixel, ds, mus);
        if (cluster == i) {
#pragma omp critical
          {
            /* sparse matrix, so need to control access, elements may be
             * rearranged */
            m(y, x) = true;
          }
        }
      }
    }
  }
  return m;
}

void ClusterModel::prepare() {
  if (locked) {
    return;
  }

  /* prepare data set for clustering */
  int originalWidth = res->getWidth();
  int originalHeight = res->getHeight();
  int pixels = originalWidth * originalHeight;
  float scale = static_cast<float>(pixels)/maxPixels;
  Image* working = res->getLow(scale);
  int workingWidth = working->getWidth();
  int workingHeight = working->getHeight();
  data.clear();

  /* prepare data set for clustering */
#pragma omp parallel
  {
    value_type pixel(3);
    std::vector<float> ds(k);
    std::vector<value_type> data1;
    int x, y;
    float s;
    data1.reserve(workingWidth * workingHeight);

#pragma omp for
    for (y = 0; y < workingHeight; ++y) {
      for (x = 0; x < workingWidth; ++x) {
        pixel(0) = working->r(y, x);
        pixel(1) = working->g(y, x);
        pixel(2) = working->b(y, x);

        /* saturation threshold */
        s = saturation(pixel(0), pixel(1), pixel(2));
        if (s >= saturationThreshold) {
          distance_type::prepare(pixel);
          data1.push_back(pixel);
        }
      }
    }

#pragma omp critical
    {
      data.insert(data.end(), data1.begin(), data1.end());
    }
  }
}

void ClusterModel::cluster() {
  const int maxIters = 100;
  if (locked) {
    return;
  }

//#pragma omp parallel
  {
    value_type pixel(3);
    std::vector<value_type> mus(k);
    vector_type ds(data.size());
    int_vector_type ks(data.size()), cs(k);
    float err, minErr = 0.0;
    int i;

//#pragma omp for
    for (i = 0; i < (int) reps; ++i) {
      clusterer.cluster(data, ks, ds, mus, cs, maxIters);
      err = ublas::norm_2(ds);
      if (i == 0 || err < minErr) {
        minErr = err;
        this->mus = mus;
      }
    }
  //}
  }

  ColourSpace cs;
  ColourSpace::rgb_t rgb(3);
  ColourSpace::hsl_t hsl(3);
  value_type pixel(3);
  for (int i = 0; i < (int) k; ++i) {
    pixel = 0.5f * mus[i] + ublas::scalar_vector<float>(3, 0.5f);
    rgb = pixel;
    cs.rgb2hsl(rgb, hsl);
    hsl(1) = 0.8f;
    hsl(2) = 0.8f;
    cs.hsl2rgb(hsl, rgb);
    colours[i].Set(uround(rgb(0)), uround(rgb(1)), uround(rgb(2)));
  }
}
